/*
 * sys-mint.c - System-dependent procedures for setting up
 * PPP interfaces on MiNT systems
 *
 * MintNet specific routines (w) 1994 Kay Roemer.
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * TODO:
 */

#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <memory.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <signal.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/route.h>

#include "iflink.h"
#include "rtent.h"
#include "mint/ppp.h"	/* ../mint/ppp.h */

#include "pppd.h"
#include "ppp.h"
#include "fsm.h"
#include "ipcp.h"

static int pppfd = -10;
static char pppdev[MAXNAMELEN];

extern int kdebugflag;
extern u_long netmask;

#define MAX_IFS		32

/* prototypes */
void die         __ARGS((int));

/*
 * SET_SA_FAMILY - set the sa_family field of a struct sockaddr,
 * if it exists.
 */

#define SET_SA_FAMILY(addr, family)			\
    memset ((char *) &(addr), '\0', sizeof(addr));	\
    addr.sa_family = (family);

/*
 * set_kdebugflag - Define the debugging level for the kernel
 */

int set_kdebugflag (int requested_level)
{
    return (1);
}

/*
 * establish_ppp - Turn the serial port into a ppp interface.
 */

void establish_ppp (void)
{
    int pid = getpid ();
    char *ttydev = ttyname (fd);
    char iname[IFNAMSIZ];
    long flags;
    short sflags;

    if (!ttydev) {
	syslog (LOG_ERR, "not a tty");
	die (1);
    }
    strcpy (iname, "ppp");
    if (if_link (ttydev, iname)) {
	syslog (LOG_ERR, "cannot link %s to PPP interface: %m", ttydev);
	die (1);
    }
    if (sscanf (iname, "ppp%d", &ifunit) != 1) {
	syslog (LOG_ERR, "invalid interface name: %s", iname);
	die (1);
    }
    sprintf (pppdev, "/dev/%s", iname);
    pppfd = open (pppdev, O_RDWR|O_NDELAY);
    if (pppfd < 0) {
	syslog (LOG_ERR, "cannot open PPP device %s: %m", pppdev);
	die (1);
    }
    if (ioctl (pppfd, PPPIOCSPGRP, &pid) < 0) {
	syslog (LOG_ERR, "cannot set pgrp for interface %s: %m", iname);
	die (1);
    }
    flags = PPPO_IP_DOWN;
    if (ioctl (pppfd, PPPIOCSFLAGS, &flags) < 0) {
	syslog (LOG_ERR, "cannot set flags on interface %s: %m", iname);
	die (1);
    }
    if (if_getifflags (iname, &sflags) ||
        if_setifflags (iname, sflags | IFF_UP)) {
	syslog (LOG_ERR, "cannot UP interface %s: %m", iname);
	die (1);
    }
}

/*
 * disestablish_ppp - Restore the serial port to normal operation.
 * This shouldn't call die() because it's called from die().
 */

void disestablish_ppp(void)
{
    short flags;

    if (pppfd >= 0) {
    	close (pppfd);
    	pppfd = -10;
	if (if_getifflags (ifname, &flags) ||
	    if_setifflags (ifname, flags | IFF_UP) ||
	    if_setifflags (ifname, flags & ~IFF_UP)) {
		syslog (LOG_WARNING, "cannot shut down interface %s", ifname);
		return;
	}
    }
}

/*
 * output - Output PPP packet.
 */

void output (int unit, unsigned char *p, int len)
{
    int r;

    if (debug)
	log_packet (p, len, "sent ");

    while ((r = write (pppfd, p, len)) == 0)
	usleep (100000L);
    if (r < 0) {
	syslog (LOG_ERR, "write: %m");
	die (1);
    }
}

/*
 * read_packet - get a PPP packet from the serial device.
 */

int read_packet (unsigned char *buf)
{
    int len;

    len = read (pppfd, buf, MTU + DLLHEADERLEN);
    if (len < 0) {
	syslog (LOG_ERR, "read: %m");
	die (1);
    } else if (len == 0)
	return -1;
    return len;
}

/*
 * ppp_send_config - configure the transmit characteristics of
 * the ppp interface.
 */

void ppp_send_config (int unit,int mtu,u_long asyncmap,int pcomp,int accomp)
{
    u_long flags, xmap[8];

    if (ioctl (pppfd, PPPIOCGFLAGS, &flags) < 0) {
	syslog (LOG_ERR, "cannot get flags on interface %s: %m", ifname);
	quit ();
    }
    flags = pcomp  ? (flags | PPPO_PROT_COMP) : (flags & ~PPPO_PROT_COMP);
    flags = accomp ? (flags | PPPO_ADDR_COMP) : (flags & ~PPPO_ADDR_COMP);
    if (ioctl (pppfd, PPPIOCSFLAGS, &flags) < 0) {
	syslog (LOG_ERR, "cannot set flags on interface %s: %m", ifname);
	quit ();
    }
    if (ioctl (pppfd, PPPIOCSMTU, &mtu) < 0) {
	syslog (LOG_ERR, "cannot set mtu on interface %s: %m", ifname);
	quit ();
    }
    if (ioctl (pppfd, PPPIOCGXMAP, xmap) < 0) {
	syslog (LOG_ERR, "cannot get xmap on interface %s: %m", ifname);
	quit ();
    }
    xmap[0] = asyncmap;
    if (ioctl (pppfd, PPPIOCSXMAP, xmap) < 0) {
	syslog (LOG_ERR, "cannot set xmap on interface %s: %m", ifname);
	quit ();
    }
}

/*
 * ppp_set_xaccm - set the extended transmit ACCM for the interface.
 */
void
ppp_set_xaccm(unit, accm)
    int unit;
    u_long *accm;
{
    MAINDEBUG ((LOG_DEBUG, "set_xaccm: %08lx %08lx %08lx %08lx\n",
		accm[0], accm[1], accm[2], accm[3]));
    if (ioctl (pppfd, PPPIOCSXMAP, accm) < 0)
	syslog (LOG_WARNING, "cannot set extended ACCM on if %s: %m", ifname);
}

/*
 * ppp_recv_config - configure the receive-side characteristics of
 * the ppp interface.
 *
 * BUG: cannot tell PPP to reject frames with proto/address compression
 * turned on.
 */

void ppp_recv_config (int unit,int mru,u_long asyncmap,int pcomp,int accomp)
{
    long flags;

    if (ioctl (pppfd, PPPIOCSMRU, &mru) < 0) {
	syslog (LOG_ERR, "cannot set mru on interface %s: %m", ifname);
	quit ();
    }
    if (ioctl (pppfd, PPPIOCSRMAP, &asyncmap) < 0) {
	syslog (LOG_ERR, "cannot set rmap on interface %s: %m", ifname);
	quit ();
    }
}

/*
 * sifvjcomp - config tcp header compression
 */

int sifvjcomp (int u, int vjcomp, int cidcomp, int maxcid)
{
    long flags;

    if (ioctl (pppfd, PPPIOCGFLAGS, &flags) < 0) {
	syslog (LOG_ERR, "cannot get flags on interface %s: %m", ifname);
	return 0;
    }
    flags |= PPPO_AUTOCOMP;
    flags = vjcomp  ? (flags | PPPO_COMPRESS) : (flags & ~PPPO_COMPRESS);
    flags = cidcomp ? (flags | PPPO_COMPCID)  : (flags & ~PPPO_COMPCID);
    if (ioctl (pppfd, PPPIOCSFLAGS, &flags) < 0) {
	syslog (LOG_ERR, "cannot set flags on interface %s: %m", ifname);
	return 0;
    }
    /*
     * The maxcid is currently hardcoded to 16 (MintNet 0.60)
     */
    flags = (long)maxcid;
    if (ioctl (pppfd, PPPIOCSSLOTS, &flags) < 0) {
	syslog (LOG_ERR, "cannot set max VJ comp slot ID: %m");
	return 0;
    }
    return 1;
}

/*
 * sifup - Config the interface up and enable IP packets to pass.
 */

int sifup (int u)
{
    long lflags;

    if (ioctl (pppfd, PPPIOCGFLAGS, &lflags) < 0) {
	syslog (LOG_ERR, "cannot get flags on interface %s: %m", ifname);
	return 0;
    }
    lflags &= ~PPPO_IP_DOWN;
    if (ioctl (pppfd, PPPIOCSFLAGS, &lflags) < 0) {
	syslog (LOG_ERR, "cannot set flags on interface %s: %m", ifname);
	return 0;
    }
    return 1;
}

/*
 * sifdown - Config the interface down and disable IP.
 */

int sifdown (int u)
{
    long lflags;

    if (ioctl (pppfd, PPPIOCGFLAGS, &lflags) < 0) {
	syslog (LOG_ERR, "cannot get flags on interface %s: %m", ifname);
	return 0;
    }
    lflags |= PPPO_IP_DOWN;
    if (ioctl (pppfd, PPPIOCSFLAGS, &lflags) < 0) {
	syslog (LOG_ERR, "cannot set flags on interface %s: %m", ifname);
	return 0;
    }
    return 1;
}

/*
 * sifaddr - Config the interface IP addresses and netmask.
 */

int sifaddr (int unit, int our_adr, int his_adr, int net_mask)
{
    struct ifreq   ifr; 
    struct rtentry rt;
    
    SET_SA_FAMILY (ifr.ifr_addr,    AF_INET); 
    SET_SA_FAMILY (ifr.ifr_dstaddr, AF_INET); 

    strncpy (ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
/*
 *  Set our IP address
 */
    ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr = our_adr;
    if (ioctl(s, SIOCSIFADDR, (caddr_t) &ifr) < 0) {
	syslog (LOG_ERR, "cannot set address on interface %s: %m", ifname);
	return 0;
    } 
/*
 *  Set the gateway address
 */
    ((struct sockaddr_in *) &ifr.ifr_dstaddr)->sin_addr.s_addr = his_adr;
    if (ioctl(s, SIOCSIFDSTADDR, (caddr_t) &ifr) < 0) {
	syslog (LOG_ERR, "cannot set dst address on interface %s: %m", ifname);
	return 0;
    }
/*
 *  Set the netmask
 */
    if (net_mask != 0) {
	((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr = net_mask;
	if (ioctl(s, SIOCSIFNETMASK, (caddr_t) &ifr) < 0) {
	    syslog (LOG_ERR, "cannot set netmask on interface %s: %m", ifname);
	    return 0;
        }
    }
/*
 *  Add the device route
 */
    memset (&rt, '\0', sizeof (rt));

    SET_SA_FAMILY (rt.rt_dst,     AF_INET);
    SET_SA_FAMILY (rt.rt_gateway, AF_INET);

    ((struct sockaddr_in *) &rt.rt_gateway)->sin_addr.s_addr = 0;
    ((struct sockaddr_in *) &rt.rt_dst)->sin_addr.s_addr     = his_adr;
    rt.rt_flags = RTF_UP | RTF_HOST;

    if (ioctl(s, SIOCADDRT, &rt) < 0) {
	syslog (LOG_ERR, "cannot add device route: %m");
	return 0;
    }
    return 1;
}

/*
 * cifaddr - Clear the interface IP addresses, and delete routes
 * through the interface if possible.
 */

int cifaddr (int unit, int our_adr, int his_adr)
{
    struct rtentry rt;
/*
 *  Delete the route through the device
 */
    memset (&rt, '\0', sizeof (rt));

    SET_SA_FAMILY (rt.rt_dst,     AF_INET);
    SET_SA_FAMILY (rt.rt_gateway, AF_INET);

    ((struct sockaddr_in *) &rt.rt_gateway)->sin_addr.s_addr = 0;
    ((struct sockaddr_in *) &rt.rt_dst)->sin_addr.s_addr     = his_adr;
    rt.rt_flags = RTF_UP | RTF_HOST;

    if (ioctl(s, SIOCDELRT, &rt) < 0) {
	syslog (LOG_ERR, "cannot delete device route: %m");
	return 0;
    }
    return 1;
}

/*
 * sifdefaultroute - assign a default route through the address given.
 */

int sifdefaultroute (int unit, int gateway)
{
    struct rtentry rt, *rtp;

    setrtent ();
    while ((rtp = getrtent ())) {
	if (rtp->rt_flags & RTF_GATEWAY &&
	    !((struct sockaddr_in *)&rtp->rt_dst)->sin_addr.s_addr) {
		syslog (LOG_WARNING, "default route already exits");
		endrtent ();
		return 0;
	}
    }
    endrtent ();

    memset (&rt, '\0', sizeof (rt));
    SET_SA_FAMILY (rt.rt_dst,     AF_INET);
    SET_SA_FAMILY (rt.rt_gateway, AF_INET);
    ((struct sockaddr_in *) &rt.rt_gateway)->sin_addr.s_addr = gateway;
    
    rt.rt_flags = RTF_UP | RTF_GATEWAY;
    if (ioctl(s, SIOCADDRT, &rt) < 0) {
	syslog (LOG_ERR, "cannot add default route: %m");
	return 0;
    }
    return 1;
}

/*
 * cifdefaultroute - delete a default route through the address given.
 */

int cifdefaultroute (int unit, int gateway)
{
    struct rtentry rt;
  
    SET_SA_FAMILY (rt.rt_dst,     AF_INET);
    SET_SA_FAMILY (rt.rt_gateway, AF_INET);
    ((struct sockaddr_in *) &rt.rt_gateway)->sin_addr.s_addr = gateway;
    
    rt.rt_flags = RTF_UP | RTF_GATEWAY;
    if (ioctl(s, SIOCDELRT, &rt) < 0) {
	syslog (LOG_ERR, "default route ioctl(SIOCDELRT): %m");
	return 0;
    }
    return 1;
}

/*
 * sifproxyarp - Make a proxy ARP entry for the peer.
 */

int sifproxyarp (int unit, u_long his_adr)
{
    struct arpreq arpreq;

    memset (&arpreq, '\0', sizeof(arpreq));
/*
 * Get the hardware address of an interface on the same subnet
 * as our local address.
 */
    if (!get_ether_addr(his_adr, &arpreq.arp_ha)) {
	syslog(LOG_ERR, "Cannot determine ethernet address for proxy ARP");
	return 0;
    }
    
    SET_SA_FAMILY(arpreq.arp_pa, AF_INET);
    ((struct sockaddr_in *) &arpreq.arp_pa)->sin_addr.s_addr = his_adr;
    arpreq.arp_flags = ATF_PERM | ATF_PUBL;
    
    if (ioctl(s, SIOCSARP, &arpreq) < 0) {
	syslog(LOG_ERR, "ioctl(SIOCSARP): %m");
	return 0;
    }
    return 1;
}

/*
 * cifproxyarp - Delete the proxy ARP entry for the peer.
 */

int cifproxyarp (int unit, u_long his_adr)
{
    struct arpreq arpreq;
  
    memset (&arpreq, '\0', sizeof(arpreq));
    SET_SA_FAMILY(arpreq.arp_pa, AF_INET);
    
    ((struct sockaddr_in *) &arpreq.arp_pa)->sin_addr.s_addr = his_adr;
    if (ioctl(s, SIOCDARP, &arpreq) < 0) {
	syslog(LOG_WARNING, "ioctl(SIOCDARP): %m");
	return 0;
    }
    return 1;
}
     
/*
 * get_ether_addr - get the hardware address of an interface on the
 * the same subnet as ipaddr.
 */

int get_ether_addr (u_long ipaddr, struct sockaddr *hwaddr)
{
    struct ifreq *ifr, *ifend, *ifp;
    int i;
    u_long ina, mask;
    struct ifreq ifreq;
    struct ifconf ifc;
    struct ifreq ifs[MAX_IFS];
    
    ifc.ifc_len = sizeof(ifs);
    ifc.ifc_req = ifs;
    if (ioctl(s, SIOCGIFCONF, &ifc) < 0) {
	syslog(LOG_ERR, "cannot get interface list: %m");
	return 0;
    }
    MAINDEBUG ((LOG_DEBUG, "proxy arp: scanning %d interfaces for IP %s",
		ifc.ifc_len / sizeof(struct ifreq), ip_ntoa(ipaddr)));
/*
 * Scan through looking for an interface with an Internet
 * address on the same subnet as `ipaddr'.
 */
    ifend = ifs + (ifc.ifc_len / sizeof(struct ifreq));
    for (ifr = ifc.ifc_req; ifr < ifend; ifr++) {
	if (ifr->ifr_addr.sa_family == AF_INET) {
	    ina = ((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr.s_addr;
	    strncpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));
            MAINDEBUG ((LOG_DEBUG, "proxy arp: examining interface %s",
			ifreq.ifr_name));
/*
 * Check that the interface is up, and not point-to-point
 * or loopback.
 */
	    if (ioctl(s, SIOCGIFFLAGS, &ifreq) < 0)
		continue;
	    if ((ifreq.ifr_flags &
		 (IFF_UP|IFF_BROADCAST|IFF_POINTOPOINT|IFF_LOOPBACK|IFF_NOARP))
		!= (IFF_UP|IFF_BROADCAST))
		continue;
/*
 * Get its netmask and check that it's on the right subnet.
 */
	    ifreq.ifr_addr.sa_family = AF_INET;
	    if (ioctl(s, SIOCGIFNETMASK, &ifreq) < 0)
	        continue;
	    mask = ((struct sockaddr_in *) &ifreq.ifr_addr)->sin_addr.s_addr;
	    MAINDEBUG ((LOG_DEBUG, "proxy arp: interface addr %s mask %lx",
			ip_ntoa(ina), ntohl(mask)));
	    if (((ipaddr ^ ina) & mask) != 0)
	        continue;
	    break;
	}
    }
    
    if (ifr >= ifend)
        return 0;

    syslog(LOG_INFO, "found interface %s for proxy arp", ifreq.ifr_name);

/*
 * Now get the hardware address.
 */
    if (ioctl (s, SIOCGIFHWADDR, &ifreq) < 0) {
        syslog(LOG_ERR, "SIOCGIFHWADDR(%s): %m", ifreq.ifr_name);
        return 0;
    }
    *hwaddr = ifreq.ifr_addr;

#define SHW(x)	((struct sockaddr_hw *)x)

    MAINDEBUG ((LOG_DEBUG,
		"proxy arp: found hwaddr %02x:%02x:%02x:%02x:%02x:%02x",
		(int) SHW(hwaddr)->shw_addr[0],
		(int) SHW(hwaddr)->shw_addr[1],
		(int) SHW(hwaddr)->shw_addr[2],
		(int) SHW(hwaddr)->shw_addr[3],
		(int) SHW(hwaddr)->shw_addr[4],
		(int) SHW(hwaddr)->shw_addr[5]));
    return 1;
}

/*
 * ppp_available - check whether the system has any ppp interfaces
 * (in fact we check whether we can do an ioctl on ppp0).
 */

int ppp_available(void)
{
    int s, ok;
    struct ifreq ifr;
    
    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s < 0)
	return 1;		/* can't tell - maybe we're not root */
    
    strncpy(ifr.ifr_name, "ppp0", sizeof (ifr.ifr_name));
    ok = ioctl(s, SIOCGIFFLAGS, (caddr_t) &ifr) >= 0;
    close(s);
    
    return ok;
}
